--# -path=.:abstract:prelude
abstract core = {
cat
	Class;	 
	Instance Class;
	Subtype (sub, super : Class);

    	Constraints;
	Constraint;
	ClassConstraintBody;
	OperConstraintBody;
	
	
	ConstraintHideContext;
	ConstraintHCInvariant;

	VarSelf (c : Class);
	VarResult (c : Class);

	InstIter (iter, exp : Class);
	SentIter (iter : Class);
	CollList Class;

	Sent;
    	AtomSent ; -- negatable sentence
	
	LetDef Class;
	LetDefSent;
	LetDefList;
	
	InstNoType;
	
	NP;

	Document;
	
    AndList;
    OrList;
    SumList;
    ProductList;
 
 --
 -- Properties
 --
    -- Properties represent the attributes/operations/associations of classes.
    -- The categories for properties are dependent types,
    -- so that they contain type information.
    -- E.g. for for a class Person with an integer attribute "age", 
    -- and a query "income(d:Date, c:Currency) : Integer", 
    -- we would get something like:
    -- Person_age_attr : SimpleProperty Person Integer;
    -- Person_income_query : Instance Date -> Instance Currency -> SimpleProperty Person Integer;
    
    SimpleProperty Class Class;
    OnProperty Class Class;    
    StaticProperty Class Class;
    AssocProperty Class Class;
   
    -- Boolean and Is properties are special cases. They always return AtomSents,
    -- so no return type information is necessary. 
    BooleanProperty Class;
    IsProperty Class;

    
fun	
    
	testNP : NP;
	dropType : (c:Class) -> Instance c -> InstNoType;

	self : (c:Class) -> VarSelf c -> Instance c;
	result : (c:Class) -> VarResult c -> Instance c;
	
	atPre : (c:Class) -> Instance c -> Instance c;
	atPreSent : Sent -> Sent;

	oneConstraint : Constraint -> Constraints;
	consConstraint : Constraint -> Constraints -> Constraints;
	document : Constraints -> Document;	
	
	defC : (c:Class) -> LetDef c -> ClassConstraintBody -> ClassConstraintBody;
	defSentC : LetDefSent -> ClassConstraintBody -> ClassConstraintBody;
	invC : Sent -> ClassConstraintBody -> ClassConstraintBody;
	invCt : Sent -> ClassConstraintBody;
	emptyClassC : ClassConstraintBody;

	emptyOperC : OperConstraintBody;
	preC : Sent -> OperConstraintBody -> OperConstraintBody;
	postC : Sent -> OperConstraintBody -> OperConstraintBody;
	prepostCt : Sent -> Sent -> OperConstraintBody;
	emptyPostC : Sent;
	
	invHC : Sent -> ConstraintHCInvariant;
 
 	subTypeRefl : (c : Class) -> Subtype c c;
 	subTypeTrans : (a, b, c : Class) -> Subtype a b -> Subtype b c -> Subtype a c;
 	coerce : (sub,super:Class) -> Subtype sub super -> Instance sub -> Instance super;
 
	eq : (c:Class) -> (a,b : Instance c) -> Sent;
	neq : (c:Class) -> (a,b : Instance c) -> Sent;
	

--
-- Let definitions
--

fun 
-- this does not work in OCL because we have no implicit propcalls there:
--     letDef : (c:Class) -> (definiendum : Instance c) -> 
--         (definiens : Instance c) -> LetDef c;
--     letDefSent : (definiendum : Sent) -> (definiens : Sent) -> LetDefSent;


letSimpleProperty : (rec,ret:Class) -> 
        (definiendum : SimpleProperty rec ret) -> 
        (definiens : Instance ret) -> 
        LetDef ret;
    letOnProperty : (rec,ret:Class) -> 
        (definiendum : OnProperty rec ret) -> 
        (definiens : Instance ret) -> 
        LetDef ret;

    letBooleanProperty : (rec:Class) -> 
        (definiendum : BooleanProperty rec) -> 
        (definiens : Sent) -> 
        LetDefSent;
    letIsProperty : (rec:Class) -> 
        (definiendum : IsProperty rec) -> 
        (definiens : Sent) -> 
        LetDefSent;

    letAddArg : (c,d:Class) -> (Instance c -> LetDef d) -> LetDef d;
    letAddArgSent : (c:Class) -> (Instance c -> LetDefSent) -> LetDefSent;

    -- to get parameter types in OCL linearization of letdefs:
    typeFormalParam : (c:Class) -> Instance c -> Instance c;

--     defC : (c:Class) -> LetDef c -> ConstraintBody -> ConstraintBody;
--     defCSent : LetDefSent -> ConstraintBody -> ConstraintBody;

    nilLet : LetDefList;
    consLet : (c:Class) -> LetDef c -> LetDefList -> LetDefList;
    consLetSent : LetDefSent -> LetDefList -> LetDefList;

    -- let only allowed as top level expr, which always has type Sent:
    localLet : LetDefList -> Sent -> Sent; 

--
-- AtomSent
--
	posAtom : AtomSent -> Sent ;
    negAtom : AtomSent -> Sent ;

--
-- Properties
--
    useSimpleProperty : (rec,ret:Class) -> SimpleProperty rec ret  
        -> Instance rec -> Instance ret;
    useSimplePropertyImplicit : (rec,ret:Class) -> SimpleProperty rec ret 
        -> Instance rec -> Instance ret;

    useBooleanProperty : (rec:Class) -> BooleanProperty rec 
        -> Instance rec -> AtomSent;
    useBooleanPropertyImplicit : (rec:Class) -> BooleanProperty rec 
        -> Instance rec -> AtomSent;

    useOnProperty : (rec,ret:Class) -> OnProperty rec ret 
        -> Instance rec -> Instance ret;
    useOnPropertyImplicit : (rec,ret:Class) -> OnProperty rec ret 
        -> Instance rec -> Instance ret;

    useIsProperty : (rec:Class) -> IsProperty rec 
        -> Instance rec -> AtomSent;
    useIsPropertyImplicit : (rec:Class) -> IsProperty rec 
        -> Instance rec -> AtomSent;

    useStaticProperty : (rec,ret:Class) -> StaticProperty rec ret 
        -> Instance rec -> Instance ret;
    useStaticPropertyImplicit : (rec,ret:Class) -> StaticProperty rec ret 
        -> Instance rec -> Instance ret;	    

    useAssocProperty : (rec,ret:Class) -> AssocProperty rec ret 
        -> Instance rec -> Instance ret;
    useAssocPropertyImplicit : (rec,ret:Class) -> AssocProperty rec ret 
        -> Instance rec -> Instance ret;





{-
--	letArg0 : (ret:Class) -> (name : String) -> Instance ret -> LetDef ret;
--	letArgCons : (argT, ret: Class) -> (argN : String) -> LetDef ret -> LetDef ret;

	

-- 	letExp2Inst : (c:Class) -> LetExp c -> Instance c;
 	letSingleAttr : (c1,c2:Class) -> Instance c1 -> (Instance c1 -> Instance c2) -> LetExp c2;
 	letConsAttr : (c1,c2:Class) -> Instance c1 -> (Instance c1 -> LetExp c2) -> LetExp c2;
	letSingleOp1arg : (c1,c2,c3:Class) -> (Instance c1 -> Instance c2) -> 
		((Instance c1 -> Instance c2) -> LetExp c3);

-- 	argSingle : (c1,c2:Class) -> (Instance c1 -> Instance c2) -> LetOpBinding c2;
-- 	argCons : (c1,c2:Class) -> (Instance c1 -> LetOpBinding c2) -> LetOpbinding c2;
 	
 	letExp2Sent :  LetExpSent -> Sent;
 	letSingleAttrS : (c1:Class) -> Instance c1 -> (Instance c1 -> Sent) -> LetExpSent;
 	letConsAttrS : (c1:Class) -> Instance c1 -> (Instance c1 -> LetExpSent) -> LetExpSent;
-} 



};
